class PaginationError(Exception):
    pass


class NoMatchingDataError(Exception):
    pass


class InvalidPSRTypeError(Exception):
    pass


class InvalidBusinessParameterError(Exception):
    pass


class InvalidParameterError(Exception):
    pass